#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Written by Mr. HuChang

import torch.nn as _nn
import torch.nn.functional as _F
from .attention import ConvMultiHeadAttention as _CMHA
import matplotlib.pyplot as _plt
import matplotlib_inline as _mat
import numpy as _np
import seaborn as _sns
import sys as _sys
from torchvision.utils import make_grid
import torch as _torch
from IPython import display as _display


_mat.backend_inline.set_matplotlib_formats("svg")


def torch_summary(net, human_style=False):
    total_num = sum(p.numel() for p in net.parameters())
    trainable_num = sum(p.numel() for p in net.parameters() if p.requires_grad)

    return {'Total': num_transform(total_num) if human_style else total_num,
            'Trainable': num_transform(trainable_num) if human_style else trainable_num,
            "Non-Trainable": num_transform(total_num - trainable_num) if human_style else (total_num - trainable_num)}


def num_transform(num):

    if num / 1000 < 1:
        return num

    elif num / 1000000 < 1:
        return str(num // 1000) + "K"

    else:
        return str(num // 1000000) + "M"



class ClassicalBlock(_nn.Module):

    def __init__(self, in_channles, out_channels, kernel_size,
                 stride, padding):
        super().__init__()
        self.backbone = _nn.Sequential(
            _nn.Conv2d(in_channles, out_channels, kernel_size,
                       stride, padding),
            _nn.BatchNorm2d(out_channels),
            _nn.ReLU(),
        )

    def forward(self, x):
        return self.backbone(x)


class BasicBlock(_nn.Module):
    """
    Example:
        >>>m = BasicBlock(3, 12, 2)
        >>>data = torch.randn(5, 3, 32, 32)
        >>>m(data).shape>>>
        >>>torch.Size([5, 12, 16, 16])
    """
    def __init__(self, in_channles, out_channels, stride):
        super().__init__()
        self.features = ClassicalBlock(in_channles, out_channels, 3,
                                       stride, 1)
        self.net = _nn.Sequential(
            _nn.Conv2d(out_channels, out_channels, 3, 1, 1),
            _nn.BatchNorm2d(out_channels)
        )

        self.shortcut = _nn.Conv2d(in_channles, out_channels, kernel_size=3,
                                   padding=1, stride=stride)

    def forward(self, x):
        y = self.features(x)
        y = self.net(y)
        z = self.shortcut(x)
        return _F.relu(y + z)


class TrainProfiler(object):
    """
    Example:
        >>>profiler = hzau.misc.TrainProfiler(len(data_loader), metric=True)
        >>>Epoch = 50
        >>>for i in range(Epoch):
        >>>    for idx, (d, l) in enumerate(data_loader):
        >>>       >>> d = d.to("cuda")
        >>>        l = l.long().to("cuda")
        >>>        out, probs = model(d)
        >>>        loss = criterion(out, l)
        >>>        optimizer.zero_grad()
        >>>        loss.backward()
        >>>        optimizer.step()
        >>>        metric = hzau.misc.cal_acc(probs, l)
        >>>        profiler.log(idx, loss, metric)
        >>>
        >>>    profiler.loss_curve()
        >>>    profiler.mertic_curve()
    """

    def __init__(self, step_length, metric=False):
        self.temp_loss = []
        self.losses = []
        self.trained_epoch = 0
        self.current_step = 0
        self.step_length = step_length
        self.metric = metric
        if metric:
            self.metrics = []
            self.temp_metric = []


    def log(self, step, loss, metric=None,
            log=True, file=_sys.stdout, clear=False):
        self.current_step += 1
        if log and metric == None:
            print("Epoch={:<3} Step[{:>3}/{}] loss={:.5f}".format(
                self.trained_epoch + 1, self.current_step,
                self.step_length, loss.item()), file=file)

        if log and metric != None:
            print("Epoch={:<3} Step[{:>3}/{}] loss={:.5f} metric={:.5f}".format(
                self.trained_epoch + 1,  self.current_step,
                self.step_length, loss.item(), metric), file=file)

        self.temp_loss.append(loss.item())
        if metric != None:
            if isinstance(metric, _torch.Tensor):
                self.temp_metric.append(metric.item())
            else:
                self.temp_metric.append(metric)

        if self.current_step == self.step_length:
            self.current_step = 0
            self.trained_epoch += 1
            self.losses.append(_np.mean(self.temp_loss))
            self.temp_loss = []
            if metric != None:
                self.metrics.append(_np.mean(self.temp_metric))
                self.temp_metric = []
        if clear == True:
            _display.clear_output()

    def loss_curve(self, x_label="x", y_label="y",
                   title="Loss Curve", delta=.00001):

        ax = _sns.lineplot(x=_np.arange(len(self.losses)) + 1, y=self.losses)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_title(title)
        _plt.show()
        _plt.pause(delta)


    def mertic_curve(self, x_label="x", y_label="y",
                   title="Metric Curve", delta=.00001):
        ax = _sns.lineplot(x=_np.arange(len(self.metrics)) + 1, y=self.metrics)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_title(title)
        _plt.show()
        _plt.pause(delta)


def imshow(data):
    images = make_grid(data.detach().cpu(), normalize=True, padding=5, pad_value=1)
    _plt.imshow(images.permute(1, 2, 0))
    _plt.axis("off")
    _plt.grid(False)
    _plt.pause(0.00001)


def cal_acc(out, labels):

    return sum(out.argmax(dim=-1) == labels) / sum(labels.shape)


class ModelProfiler(TrainProfiler):
    """
    Example:
        >>>profiler = hzau.misc.ModelProfiler(len(data_loader), metric=True)
        >>>Epoch = 100
        >>>model = model.train()
        >>>for i in range(Epoch):
        >>>    for idx, (d, l) in enumerate(data_loader):
        >>>        d = d.to(>>>"cuda")
        >>>        l = l.long().to("cuda")
        >>>        out, probs = model(d)
        >>>        loss = criterion(out, l)
        >>>        optimizer.zero_grad()
        >>>        loss.backward()
        >>>        optimizer.step()
        >>>        metric = hzau.misc.cal_acc(probs, l)
        >>>        if profiler.trained_epoch % 20 == 0 and profiler.current_step == 1:
        >>>            profiler.log(idx, loss, metric, clear=True)
        >>>        profiler.log(idx, loss, metric)
        >>>    loss, acc = evaluation(model)
        >>>    profiler.log_eval(loss, acc)
        >>>    model = model.train()
        >>>    scheduler.step()
        >>>    profiler.loss_curve()
        >>>    profiler.mertic_curve()
    """
    def __init__(self, step_length, metric=False):

        super().__init__(step_length, metric)
        self.eval_losses = []
        self.eval_metric = []

    def log_eval(self, loss, metric):
        if isinstance(loss, _torch.Tensor):
            self.eval_losses.append(loss.item())
        else:
            self.eval_losses.append(loss)
        if isinstance(metric, _torch.Tensor):
            self.eval_metric.append(metric.item())
        else:
            self.eval_metric.append(metric)

    def loss_curve(self, x_label="x", y_label="y",
                   title="Loss Curve", delta=.00001):

        ax = _sns.lineplot(x=_np.arange(len(self.losses)) + 1,
                           y=self.losses, label="Train")
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_title(title)
        ax.plot(_np.arange(len(self.eval_losses)) + 1,
                self.eval_losses, label="Test")
        ax.legend()
        _plt.show()
        _plt.pause(delta)


    def mertic_curve(self, x_label="x", y_label="y",
                   title="Metric Curve", delta=.00001):
        ax = _sns.lineplot(x=_np.arange(len(self.metrics)) + 1,
                           y=self.metrics, label="Train")
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_title(title)
        ax.plot(_np.arange(len(self.eval_metric)) + 1,
                self.eval_metric, label="Test")
        ax.legend()
        _plt.show()
        _plt.pause(delta)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Written by Mr. HuChang


from __future__ import print_function
import torch.nn as _nn
import torch.nn.functional as _F
import math as _math
import torch as _torch


class SelfAttention(_nn.Module):
    """
    Input shape: B x C x F
    apply self attention in dimension `C`
    Parameters:
        W: dimension of the column of the matrix
        d_k: same size as C
        output_size: the output size of the last dimension
        verbose: print attention matrix dimension
    Example:
        >>>data = torch.randn(5, 625, 3)
        >>>m = SelfAttention(W=3, d_k=100, output_size=100)
        >>>m(data).shape
        >>>torch.Size([5, 625, 100])
        ============================
        >>>data = torch.randn(5, 625, 3)
        >>>m = SelfAttention(W=3, d_k=100)
        >>>m(data).shape
        >>>torch.Size([5, 625, 3])
        ============================
        >>>data = torch.randn(5, 3, 32, 32)
        >>>data = data.view(5, 3, -1)
        >>>data = data.transpose(1, 2)
        >>>m = SelfAttention(W=3, d_k=100)
        >>>m(data).shape
        >>>torch.Size([5, 1024, 3])
        ===========================
        >>>data = torch.randn(5, 3, 32, 32)
        >>>data = data.view(5, 3, -1)
        >>>data = data.transpose(1, 2)
        >>>m = SelfAttention(W=3, d_k=100, output_size=1000, verbose=True)
        >>>the attention matrix shape is torch.Size([5, 1024, 1024])
        >>>torch.Size([5, 1024, 1000])
    """

    def __init__(self, W, d_k=None, output_size=None, dropout=0.,
                 verbose=False):
        super().__init__()
        if output_size == None:
            output_size = W
        if d_k == None:
            d_k = W
        self.Q = _nn.Linear(W, d_k)
        self.K = _nn.Linear(W, d_k)
        self.V = _nn.Linear(W, output_size)
        self.d_k = d_k
        self.dropout = _nn.Dropout(dropout)
        self.verbose = verbose

    def forward(self, x):
        assert len(x.shape) == 3, "x's shape should be B x C x F"
        Q = self.Q(x)
        K = self.K(x)
        V = self.V(x)
        dot_product = _torch.bmm(Q, K.transpose(1, 2)) / _math.sqrt(self.d_k)
        dot_product = _F.softmax(dot_product, dim=-1)
        if self.verbose:
            print(f"the attention matrix shape is {dot_product.shape}")
        out = self.dropout(dot_product)
        out = _torch.bmm(out, V)
        return out


class MultiHeadAttention(_nn.Module):
    """
    Input shape: B x C x F
    apply self attention in dimension `C`
    Parameters:
        W: dimension of the column of the matrix
        num_heads: number of heads
        d_k: same as Self Attention default = W
        output_size: the output size of the last dimension default = W
        verbose: print attention matrix dimension
        V_outsize: Value matrix column size
    Example:
        >>>m = MultiHeadAttention(100, 10, output_size=200, verbose=True)
        >>>data = torch.randn(5, 50, 100)
        >>>m(data).shape
        >>>the attention matrix shape is torch.Size([5, 10, 50, 50])
        >>>torch.Size([5, 50, 200])
        ====================================
        >>>m = MultiHeadAttention(100, 5, verbose=True)
        >>>data = torch.randn(5, 50, 100)
        >>>m(data).shape
        >>>the attention matrix shape is torch.Size([5, 5, 50, 50])
        >>>torch.Size([5, 50, 100])
        ====================================
        >>>m = MultiHeadAttention(100, 3, V_outsize=500, verbose=True)
        >>>data = torch.randn(5, 50, 100)
        >>>m(data).shape
        >>>the attention matrix shape is torch.Size([5, 3, 50, 50])
        >>>torch.Size([5, 50, 100])
    """

    def __init__(self, W, num_heads, output_size=None, d_k=None,
                 V_outsize=None, dropout=0., verbose=False):
        super().__init__()
        self.num_heads = num_heads
        if V_outsize == None:
            V_outsize = W
        if d_k == None:
            d_k = W
        if output_size == None:
            output_size = W
        self.V_outsize = V_outsize
        self.Q = _nn.Linear(W, d_k * num_heads)
        self.K = _nn.Linear(W, d_k * num_heads)
        self.V = _nn.Linear(W, V_outsize * num_heads)
        self.d_k = d_k
        self.dropout = _nn.Dropout(dropout)
        self.verbose = verbose
        self.output_size = output_size
        self.dim_reduction = _nn.Linear(V_outsize * num_heads, output_size)

    def forward(self, x):
        assert len(x.shape) == 3, "x's shape should be B x C x F"
        B, C, f = x.shape
        Q = self.Q(x).view(B, C, -1, self.num_heads).permute(0, 3, 1, 2)
        K = self.K(x).view(B, C, -1, self.num_heads).permute(0, 3, 2, 1)
        V = self.V(x).view(B, C, -1, self.num_heads).permute(0, 3, 1, 2)
        attention = _F.softmax(_torch.matmul(Q, K) / _math.sqrt(self.d_k), -1)
        if self.verbose:
            print(f"the attention matrix shape is {attention.shape}")
        out = _torch.matmul(attention, V).view(B, C,
                                              self.V_outsize * self.num_heads)
        return self.dim_reduction(out)


class ConvSelfAttention(_nn.Module):
    """
    Input: B x C x H x W
    Parameters:
        in_channels: equals to C
        out_channels: same as out_chaneels in Conv2d
        kernel_size: size of convolutional kernel
    Example:
        >>>m = ConvSelfAttention(3, d_k=10, out_channels=12, verbose=True)
        >>>data = torch.randn(5, 3, 32, 32)
        >>>m(data).shape
        >>>the attention matrix shape is torch.Size([5, 1024, 1024])
        >>>torch.Size([5, 12, 32, 32])
    """
    def __init__(self, in_channels, d_k, out_channels, kernel_size=3,
                 verbose=False):
        super().__init__()

        self.d_k = d_k
        self.out_channels = out_channels
        self.verbos = verbose
        self.Q = _nn.Conv2d(in_channels, d_k, kernel_size=kernel_size,
                           padding="same")
        self.K = _nn.Conv2d(in_channels, d_k, kernel_size=kernel_size,
                           padding="same")
        self.V = _nn.Conv2d(in_channels, out_channels, kernel_size,
                           padding="same")

    def forward(self, x):
        assert len(x.shape) == 4, "x's shape should be B x C x H x W"
        B, C, H, W = x.shape
        Q = self.Q(x).view(B, self.d_k, -1)
        K = self.K(x).view(B, self.d_k, -1)
        V = self.V(x).view(B, self.out_channels, -1)

        attention = _F.softmax(_torch.bmm(Q.transpose(1, 2), K) / _math.sqrt(self.d_k),
                              dim=-1)
        if self.verbos:
            print(f"the attention matrix shape is {attention.shape}")
        out = _torch.bmm(V, attention.transpose(1, 2))

        return out.view(B, -1, H, W)


class ConvMultiHeadAttention(_nn.Module):
    """
    Input: B x C x H x W
    Parameters:
        in_channels: equals to C
        out_channels: same as out_chaneels in Conv2d
        kernel_size: size of convolutional kernel
        num_heads: number of self attention
        V_out_channels: Value matrix column size
    Example:
        >>>m = ConvMultiHeadAttention(in_channels=3, d_k=12, V_out_channels=12,
        >>>                          num_heads=5, out_channels=12, verbose=True)
        >>>data = torch.randn(5, 3, 32, 32)
        >>>m(data).shape
        >>>the attention matrix shape is torch.Size([5, 5, 1024, 1024])
        >>>torch.Size([5, 12, 32, 32])
    """

    def __init__(self, in_channels, d_k, V_out_channels, num_heads,
                 out_channels, kernel_size=3, verbose=False,
                 transform_dropout=.0, dropout=.0):
        super().__init__()

        self.d_k = d_k
        self.out_channels = out_channels
        self.V_out_channels = V_out_channels
        self.num_heads = num_heads
        self.verbos = verbose
        self.Q = _nn.Conv2d(in_channels, d_k * num_heads, kernel_size,
                           padding="same")
        self.K = _nn.Conv2d(in_channels, d_k * num_heads, kernel_size,
                           padding="same")
        self.V = _nn.Conv2d(in_channels, V_out_channels * num_heads, kernel_size,
                           padding="same")
        self.channel_transform = _nn.Linear(V_out_channels * num_heads,
                                           out_channels)
        self.dropout = _nn.Dropout(dropout)
        self.Q_dropout = _nn.Dropout(transform_dropout)
        self.K_dropout = _nn.Dropout(transform_dropout)
        self.V_dropout = _nn.Dropout(transform_dropout)

    def forward(self, x):
        assert len(x.shape) == 4, "x's shape should be B x C x H x W"
        B, C, H, W = x.shape
        Q = self.Q(x).view(B, self.d_k, H * W,
                           self.num_heads).permute(0, 3, 1, 2)
        K = self.K(x).view(B, self.d_k, H * W,
                           self.num_heads).permute(0, 3, 2, 1)
        V = self.V(x).view(B, self.V_out_channels, H * W,
                           self.num_heads).permute(0, 3, 1, 2)
        Q = self.Q_dropout(Q)
        V = self.V_dropout(V)
        K = self.V_dropout(K)
        attention = _F.softmax(_torch.matmul(K, Q) / _math.sqrt(self.d_k),
                              dim=-1)
        attention = self.dropout(attention)
        if self.verbos:
            print(f"the attention matrix shape is {attention.shape}")
        out = _torch.matmul(V, attention.transpose(-1, -2))
        out = self.channel_transform(out.view(B, H * W, -1))

        return out.view(B, -1, H, W)




#!/usr/bin/env python
# coding: utf-8

from setuptools import setup

setup(
    name='hzau',
    version='0.0.3',
    author='Chang-LeHung',
    author_email='chang-lehung@gmail.com',
    url='https://github/Chang-LeHung',
    description=u'utis',
    packages=['hzau'],
    install_requires=['torch', 'matplotlib', 'numpy', 'seaborn',
                      'torchvision', 'IPython', 'matplotlib_inline']
)